@@ -23,7 +23,7 @@
 //! - Because UDP is a "connectionless" protocol, it offers no guarantee of "conversation" between traditional client and server roles. All the UDP transport layer gives you is a method to listen for messages thrown at you, and to throw messages at someone. Owing to this, CoAP machines are expected to perform both client and server roles (or more accurately, _sender_ and _receiver_ roles)
 //! - While _classes_ of status codes are the same (Success 2xx -> 2.xx, Client error 4xx -> 4.xx, Server error 5xx -> 5.xx), the semantics of the individual response codes differ.
 
-#![doc(html_root_url = "https://docs.rs/kwap/0.1.10")]
+#![doc(html_root_url = "https://docs.rs/kwap/0.2.0")]
 #![cfg_attr(all(not(test), feature = "no_std"), no_std)]
 #![cfg_attr(not(test),
             deny(missing_debug_implementations,
@@ -61,6 +61,7 @@ pub mod socket;
 pub mod std;
 
 static mut ID: u16 = 0;
+static mut TOKEN: u64 = 0;
 
 fn generate_id() -> kwap_msg::Id {
   // TEMPORARY
@@ -72,6 +73,16 @@ fn generate_id() -> kwap_msg::Id {
   }
 }
 
+fn generate_token() -> kwap_msg::Token {
+  // TEMPORARY
+  // TODO: replace with long-living Client or Endpoint structure
+  #[allow(unsafe_code)]
+  unsafe {
+    TOKEN += 1;
+    kwap_msg::Token(TOKEN.to_be_bytes().into())
+  }
+}
+
 fn add_option<A: Array<Item = (OptNumber, Opt<B>)>, B: Array<Item = u8>, V: IntoIterator<Item = u8>>(
   opts: &mut A,
   number: u32,
@@ -160,8 +171,9 @@ pub(crate) mod test {
              tx: Default::default() }
     }
 
-    pub fn init(rx: Vec<u8>) -> Self {
+    pub fn init(addr: SocketAddr, rx: Vec<u8>) -> Self {
       let mut me = Self::new();
+      me.addr = Some(addr);
       me.rx = rx;
       me
     }
@@ -175,9 +187,9 @@ pub(crate) mod test {
       Ok(())
     }
 
-    fn recv(&self, buf: &mut [u8]) -> nb::Result<usize, Self::Error> {
-      if self.rx.is_empty() {
-        eprintln!("TubeSock recv invoked without any bytes");
+    fn recv(&self, buf: &mut [u8]) -> nb::Result<(usize, SocketAddr), Self::Error> {
+      if self.addr.is_none() || self.rx.is_empty() {
+        println!("TubeSock recv invoked without sending first");
         return Err(nb::Error::WouldBlock);
       }
 
@@ -190,7 +202,8 @@ pub(crate) mod test {
            .enumerate()
            .for_each(|(ix, el)| buf[ix] = el);
       }
-      Ok(n)
+
+      Ok((n, self.addr.clone().unwrap()))
     }
 
     fn send(&self, buf: &[u8]) -> nb::Result<(), Self::Error> {
